/*
 *	dratest.d -- part of the drat! packrat parser package pre-alpha v.0.20
 *
 *	Copyright (C) 2007 Frank F. Hirsch "0ffh"
 *                     (email frank at frankhirsch fullstop net)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be Tful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
module dratest;

import std.conv;
import std.date;
import std.file;
import std.stdio;
import std.regexp;
import std.string;
import std.boxer;
import dratlib;

alias Box T;
alias TDratParser!(T)  DratParser;
alias TRule!(T)        Rule;
alias TParseNode!(T)   PN;

version (D_Version2)
{
	mixin("alias const(char)* ccptr;");
	alias char* cptr;
}
else
{
	alias char*  ccptr;
	alias char*  cptr;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- default rule function set for box
//--------------------

void useBoxRuleFuns(TDratParser!(Box) p)
{
  p.defRuleFun[RT.C]=&boxRuleFuns.funC;
  p.defRuleFun[RT.S]=&boxRuleFuns.funS;
  p.defRuleFun[RT.R]=&boxRuleFuns.funR;
  p.defRuleFun[RT.E]=&boxRuleFuns.funE;
  p.defRuleFun[RT.RN]=&boxRuleFuns.funR;
  p.defRuleFun[RT.FST]=&boxRuleFuns.funFST;
  p.defRuleFun[RT.SEQ]=&boxRuleFuns.funSEQ;
  p.defRuleFun[RT.OOM]=&boxRuleFuns.funOOM;
  p.defRuleFun[RT.ZOM]=&boxRuleFuns.funZOM;
  p.defRuleFun[RT.OPT]=&boxRuleFuns.funOPT;
  p.defRuleFun[RT.AND]=&boxRuleFuns.funAND;
  p.defRuleFun[RT.NOT]=&boxRuleFuns.funNOT;
}

struct boxRuleFuns
{
  alias Box T;
  alias TParseNode!(T) PN;
  
  static T funC(PN* n)
  {
    dpf2("C:%s\n",tsz(n.match()));
    return box(n.match()[0]);
  };
  static T funS(PN* n)
  {
    dpf2("S:%s\n",tsz(n.match()));
    return box(n.match());
  };
  static T funE(PN* n)
  {
    dpf2("E:%s\n",tsz(n.match()));
    return box(n.match());
  };
  static T funR(PN* n)
  {
    dpf2("R:%s\n",tsz(n.match()));
    return n.l.call();
  };
  static T funFST(PN* n)
  {
    dpf2("FST:%s\n",tsz(n.match()));
    return n.l.call();
  };
  static T funSEQ(PN* n)
  {
    dpf2("SEQ:%s\n",tsz(n.match()));
    return box(n.subcalls());
  };
  static T funOOM(PN* n)
  {
    dpf2("OOM:%s\n",tsz(n.match()));
    return box(n.subcalls());
  };
  static T funZOM(PN* n)
  {
    dpf2("ZOM:%s\n",tsz(n.match()));
    return box(n.subcalls());
  };
  static T funOPT(PN* n)
  {
    dpf2("OPT:%s\n",tsz(n.match()));
    if (n.l)
      return n.l.call();
    else
      return box(null);
  };
  static T funAND(PN* n)
  {
    dpf2("AND:%s\n",tsz(n.match()));
    return box(null);
  };
  static T funNOT(PN* n)
  {
    dpf2("NOT:%s\n",tsz(n.match()));
    return box(null);
  };
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- test rule parser
//--------------------

Rule* testRuleParser(string src)
// implements a rule parser like the one built into dratlib.DratParser
{
	DratParser p=new DratParser();
  useBoxRuleFuns(p);
  // fst
  p.addRule(`fst:(ws seq ($[|/]$ ws seq)*)`,
    function Box(PN* n)
    { // uses "collect" to get all nodes in subtree called "seq".
      // the collect function recurses into the parse tree, but
      // does not recurse into nodes belonging to a named rule.
      // if this rule would be
      //   `fst:(ws head:seq ("/" ws seq)*)`
      //                        only ^^^ would be collected.
      //   as the name "head" would terminate recursion
      // if this rule would be
      //   `fst:(ws seq tail:("/" ws seq)*)`
      //       only ^^^ would be collected.
      //   as name "tail" would terminate recursion
      PN*[] nl=n.collect("seq");
      // more than one option?
      if (nl.length>1)
      { // if there is more then one option
        // get the resulting rules from the "seq"s
        Rule*[] rl;
        rl.length=nl.length;
        for (int k=0;k<nl.length;++k)
          // all data is by default wrapped into "Box"es,
          // but to retain flexibility the function calls
          // return void pointers, so either you use tons
          // of casts to be able to use std.Boxer directly,
          // or you just mixin the VoidBoxer template :-)))
          rl[k]=unbox!(Rule*)(nl[k].call());
        // and wrap them into a new rule of type FST (=first match)
        return box(Rule.newFST(rl));
      }
      else
        // if there is only one option
        // pass through the rule generated by "seq"
        return nl[0].call();
    });
  // seq
  p.addRule("seq:named+",
    function Box(PN* n)
    { // uses "collectCall" to get all "named"s
      // collectCall collects the call results instead of the nodes
      // the rest is analogue to the rule "fst" above
      T[] dl=n.collectCall("named");
      if (dl.length>1)
      {
        Rule*[] rl;
        rl.length=dl.length;
        for (int k=0;k<rl.length;++k)
          rl[k]=unbox!(Rule*)(dl[k]);
        // wrap the rules into a new SEQ (=match sequence) rule 
        return box(Rule.newSEQ(rl));
      }
      else
        return dl[0];
    });
  // named
  p.addRule(`named:((id:$[_a-zA-Z][-_a-zA-Z0-9]*:$)? ws decor ws)`,
    function Box(PN* n)
    {
      T d;
      // get data from decor
      // "findCall" is similar to collectCall, but only returns
      // the call result of the first node that is found
      d=n.findCall("decor");
      // "id" defined?
      if (PN* sn=n.find("id"))
      { // this rule has a name
        string s=sn.match()[0..$-1]; // strip the ":"
        unbox!(Rule*)(d).rn=s; // set rule name
      }
      return d;
    });
  // decor
  p.addRule(`decor:(prefix:$[&!]?$ undec postfix:$[?*+]?$ ws)`,
    function Box(PN* n)
    {
      Rule* r=unbox!(Rule*)(n.findCall("undec"));
      string s;
      s=n.findMatch("postfix");
      if (s=="?")
        // wrap the rule into an OPT (=optional) rule
        r=Rule.newOPT(r);
      else if (s=="*")
        // wrap the rule into an ZOM (=zero-or-more repetitions) rule
        r=Rule.newZOM(r);
      else if (s=="+")
        // wrap the rule into an OOM (=one-or-more repetitions) rule
        r=Rule.newOOM(r);
      s=n.findMatch("prefix");
      if (s=="&")
        // wrap the rule into an AND (=non-consuming and) rule
        r=Rule.newAND(r);
      else if (s=="!")
        // wrap the rule into an NOT (=non-consuming not) rule
        r=Rule.newNOT(r);
      return box(r);
    });
  // undec
  p.addRule(`undec:(atom / "(" ws fst ")" ws)`,
    function Box(PN* n)
    {
      if (PN* sn=n.find("atom"))
        // subnode atom defined
        return sn.call();
      else
        // otherwise fst must be defined
        return n.findCall("fst");
    });
  // atom
  p.addRule(`atom:(string/re/rule)`);
  // rule
  p.addRule(`rule:(expr:$[_a-zA-Z][_a-zA-Z0-9]*$ ws)`,
    function Box(PN* n)
    {
      string s=n.findMatch("expr");
      return box(Rule.newR(s));
    });
  // string
  p.addRule(`string:(expr:($'([^']|(\\'))*'$ / %"([^"]|(\\"))*"%) ws)`,
    function Box(PN* n)
    {
      string s=n.findMatch("expr")[1..$-1];
      if (s.length==1)
        // optimise string of length one to chars
        return box(Rule.newC(s[0]));
      else
        return box(Rule.newS(s));
    });
  // regexp
  p.addRule(`re:(("$" expr:%([^$]|(\$\$))*% "$" ws)/`~
                `("%" expr:$([^%]|(\%\%))*$ "%" ws))`,
    function Box(PN* n)
    {
      return box(Rule.newE(n.findMatch("expr")));
    });
  // ws
  p.addRule(`ws:($\s+$ / $#[^\r\n]*$)*`);
  // test
  PN* pt=p.parse("fst",src);
  if (pt)
    return unbox!(Rule*)(pt.call());
  else
    return null;
}

void testTestRuleParser()
// compares
//  - the testRuleParser from above
//  - the built-in rule parser of dratlib
//  - and the hand-constructed rule
{
  Rule *r;
  string src=`testRule:(option-1: (ws foo ("/" ws foo)*) | option-2:!"bar")`;
  printf("rule source is '%s'\n",std.string.toStringz(src));
  r=testRuleParser(src);
  printf("----- test rule parser\n");
  if (r)
    r.show();
  printf("----- built-in rule parser\n");
  DratParser.parseRule(src).show();
  printf("----- manual rule construction\n");
  Rule.newFST("testRule",[
    Rule.newSEQ("option-1",[
      Rule.newR("ws"),
      Rule.newR("foo"),
      Rule.newZOM(
        Rule.newSEQ([
          Rule.newC('/'),
          Rule.newR("ws"),
          Rule.newR("foo")]))]),
    Rule.newNOT("option-2",
      Rule.newS("bar"))]).show();
}

void main()
{
  testTestRuleParser();
}



