/*
 *	dratlib.d -- part of the drat! packrat parser package pre-alpha v.0.20
 *
 *	Copyright (C) 2007 Frank F. Hirsch "0ffh"
 *                     (email frank at frankhirsch fullstop net)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

module dratlib;

import std.gc;
import std.date;
import std.string;
import std.regexp;

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- D version independence
//--------------------

version (D_Version2)
{
	mixin("alias const(char)* ccptr;");
	alias char* cptr;
}
else
{
	alias char*  ccptr;
	alias char*  cptr;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- lazyness
//--------------------

alias string str;
alias std.string.toStringz tsz;

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- debugging
//--------------------

const bool debugFlag=!true;

void dummy(...)
{
}

static if (debugFlag)
	alias printf dpf;
else
	alias dummy dpf;

static if (!true)
  alias printf dpf2;
else
  alias dpf dpf2;
  

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//--------------------
//--------------------

str spaces(int n)
{
  str s=cast(str)"";
  while (n--) s~=" ";
  return s;
}

str deCtrl(str s)
{
  str hex="0123456789abcdef";
  for (int k=0;k<s.length;)
  {
    if (s[k]<0x20)
    {
      str x;
      if (k>0)
        x~=s[0..k];
      x~="\\"~hex[s[k]>>4]~hex[s[k]&15];
      if (k<s.length)
        x~=s[k+1..$];
      s=x;
    }
    else
    {
      ++k;
    }
  }
  if (s.length>25)
    return s[0..25]~"...";
  else
    return s[0..$];
}

void* myalloc(size_t s)
{
  static if (false)
    // ? not garbage collected
    return std.gc.malloc(s).ptr;
  else
    // because this works fine
    return (new void[s]).ptr;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- parsing rule types
//--------------------

/// Used to distinguish different types of rules.
enum RuleType
{
  C,      /// character
  S,      /// str 
  E,      /// regexp
  R,      /// rule
  RN,     /// rule by name
  FST,    /// first in list
  SEQ,    /// all in list (in order)
  OOM,    /// one or more
  ZOM,    /// zero or more
  OPT,    /// optional
  AND,    /// and
  NOT     /// not
};

alias RuleType RT;

str typeStr(RT t)
{
  switch (t)
  {
    case RT.C   : return "C";
    case RT.S   : return "S";
    case RT.E   : return "E";
    case RT.R   : return "R";
    case RT.RN  : return "RN";
    case RT.FST : return "FST";
    case RT.SEQ : return "SEQ";
    case RT.OOM : return "OOM";
    case RT.ZOM : return "ZOM";
    case RT.OPT : return "OPT";
    case RT.AND : return "AND";
    case RT.NOT : return "NOT";
    default     : return "n/a";
  }
}	

/// Rule
struct TRule(T)
{
  alias T function(TParseNode!(T)*) RuleFun;
  alias TParseNode!(T)*[] Memo;
  alias TRule!(T) Rule;
  alias TDratParser!(T) DP;

  RT      rt; /// type
  str	    rn; /// rule name
  bool		mf; /// memo flag
  Memo  	me; /// memo entry
  RuleFun fu; /// rule function
  str*    ps; /// associated parser's source
  DP      dp; /// associated parser object
  union
  {
    char     c; /// character (C)
    str      s; /// str (S)
    RegExp   e; /// regexp (E)
    Rule*    r; /// rule (R,OOM,ZOM,OPT,AND,NOT)
    str*     n; /// rule name (RN)
    Rule*[]  l; /// list for comprehension operators (FST, SEQ)
  }
  void show(int ind=0)
  {
    void pri(int ind)
    {
      while (ind--)
        printf(" ");
    }
    pri(ind);
    printf("%s:",tsz(this.rn));
    switch (this.rt)
    {
      case RT.C   :
      { // match character
        printf("C'%c' fu=%p\n",this.c,this.fu);
        break;
      }
      case RT.S   :
      { // match str
        printf("S'%s' fu=%p\n",tsz(this.s),this.fu);
        break;
      }
      case RT.E   :
      { // match regexp
        printf("E fu=%p\n",this.fu);
        break;
      }
      case RT.FST :
      { // match first
        printf("FST fu=%p\n",this.fu);
        foreach (lr;this.l)
          lr.show(ind+2);
        break;
      }
      case RT.SEQ :
      { // match all (serial)
        printf("SEQ fu=%p\n",this.fu);
        foreach (lr;this.l)
          lr.show(ind+2);
        break;
      }
      case RT.OPT :
      { // option
        printf("OPT fu=%p\n",this.fu);
        this.r.show(ind+2);
        break;
      }
      case RT.ZOM :
      { // zero-or-more
        printf("ZOM fu=%p\n",this.fu);
        this.r.show(ind+2);
        break;
      }
      case RT.OOM :
      { // one-or-more
        printf("OOM fu=%p\n",this.fu);
        this.r.show(ind+2);
        break;
      }
      case RT.AND :
      { // match this, don't consume input
        printf("AND fu=%p\n",this.fu);
        this.r.show(ind+2);
        break;
      }
      case RT.NOT :
      { // match all but this, don't consume input
        printf("NOT fu=%p\n",this.fu);
        this.r.show(ind+2);
        break;
      }
      case RT.R :
      { //
        printf("R'%s' fu=%p\n",tsz(this.r.rn),this.fu);
        //printf("RN\n");
        break;
      }
      case RT.RN :
      { //
        printf("RN'%s' fu=%p\n",tsz(*this.n),this.fu);
        //printf("RN\n");
        break;
      }
      default :
      { // unknown
        printf("???(%s:%i) fu=%p\n",tsz(this.rn),this.rt,this.fu);
        throw new Exception("rule.show: unsupportet rule type");
      }
    }
  }
  static Rule* newC(str rn,char c,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.C;
    r.c=c;
    r.fu=fun;
    return r;
  }
  static Rule* newS(str rn,str s,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.S;
    r.s=s~"\0"; // ! currently needed for acceptStr-function
    r.fu=fun;
    return r;
  }
  static Rule* newE(str rn,str s,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.E;
    r.e=RegExp("^"~s);
    r.fu=fun;
    return r;
  }
  static Rule* newR(str rn,str s,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.RN;
    r.n=(new str[1]).ptr;
    *(r.n)=s;
    r.fu=fun;
    return r;
  }
  static Rule* newFST(str rn,Rule*[] l,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.FST;
    r.l=l;
    r.fu=fun;
    return r;
  }
  static Rule* newSEQ(str rn,Rule*[] l,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.SEQ;
    r.l=l;
    r.fu=fun;
    return r;
  }
  static Rule* newOOM(str rn,Rule* ru,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.OOM;
    r.r=ru;
    r.fu=fun;
    return r;
  }
  static Rule* newZOM(str rn,Rule* ru,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.ZOM;
    r.r=ru;
    r.fu=fun;
    return r;
  }
  static Rule* newOPT(str rn,Rule* ru,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.OPT;
    r.r=ru;
    r.fu=fun;
    return r;
  }
  static Rule* newAND(str rn,Rule* ru,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.AND;
    r.r=ru;
    r.fu=fun;
    return r;
  }
  static Rule* newNOT(str rn,Rule* ru,RuleFun fun=null)
  {
    Rule* r=cast(Rule*)myalloc((Rule).sizeof);
    r.rn=rn;
    r.rt=RT.NOT;
    r.r=ru;
    r.fu=fun;
    return r;
  }
  static Rule* newC(char c)
  {
    return newC("",c);
  }
  static Rule* newS(str s)
  {
    return newS("",s);
  }
  static Rule* newE(str s)
  {
    return newE("",s);
  }
  static Rule* newR(str s)
  {
    return newR("",s);
  }
  /*static Rule* newW(str s)
  {
    return newW("",s);
  }*/
  static Rule* newFST(Rule*[] l)
  {
    return newFST("",l);
  }
  static Rule* newSEQ(Rule*[] l)
  {
    return newSEQ("",l);
  }
  static Rule* newOOM(Rule* r)
  {
    return newOOM("",r);
  }
  static Rule* newZOM(Rule* r)
  {
    return newZOM("",r);
  }
  static Rule* newOPT(Rule* r)
  {
    return newOPT("",r);
  }
  static Rule* newAND(Rule* r)
  {
    return newAND("",r);
  }
  static Rule* newNOT(Rule* r)
  {
    return newNOT("",r);
  }
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- types for parsing
//--------------------

/// Parse node.
struct TParseNode(T)
{
  alias TRule!(T) Rule;
  alias TParseNode!(T) PN;

  PN*     n;  /// next in list
  PN*     l;  /// sublist
  Rule*   r;  /// matched rule
  int    sb;  /// source match begin
  int    se;  /// source match end+1

  PN* opIndex(size_t idx)
  // warn in doc: syntax looks O(1) but is O(n)
  {
    return this.at(idx);
  }
  PN* opIndex(str key)
  // warn in doc: syntax looks O(1) but is O(n)
  {
    return this.at(key);
  }
  T opCall()
  {
    return call();
  }
  T call()
  {
//printf("call r.fu=%p\n",r.fu);
    return r.fu(this);
  }
  T[] subcalls()
  {
    T[] ld;
    PN* n=this.l;
    int p;
    ld.length=8;
    while (n)
    {
      if (p>=ld.length)
        ld.length=ld.length*2;
      ld[p++]=n.call();
      n=n.n;
    }
    ld.length=p;
    return ld;
  }
  PN* at(size_t idx)
  {
    PN* n=this.l;
    while (idx--)
    {
      if ((n=n.n) is null)
        return null;
    }
    return n;
  }
  PN* at(str key)
  {
    PN* n=this.l;
    while (n)
    {
      if (n.r.rn==key)
        return n;
      n=n.n;
    }
    return n; // null
  }
  PN*[] asArray()
  {
    PN* n=l;
    PN*[] nl;
    while (n)
    {
      nl~=n;
      n=n.n;
    }
    return nl;
  }
  str match()
  {
    return (*(r.ps))[sb..se];
  }
  cptr srcSamp()
  {
    return tsz(deCtrl(match()));
  }
  PN* find(str key)
  {
    PN* n=this.l;
    while (n)
    {
      if (n.r.rn==key)
        return n;
      else
        if (n.r.rn=="")
          if (PN* c=n.find(key))
            return c;
      n=n.n;
    }
    return n; // null
  }
  T findCall(str key)
  {
    PN* n=this.l;
    while (n)
    {
      if (n.r.rn==key)
        return n.call();
      else
        if (n.r.rn=="")
          if (PN* c=n.find(key))
            return c.call();
      n=n.n;
    }
    return T.init;
  }
  str findMatch(str key)
  {
    PN* n=this.l;
    while (n)
    {
      if (n.r.rn==key)
        return n.match();
      else
        if (n.r.rn=="")
          if (PN* c=n.find(key))
            return c.match();
      n=n.n;
    }
    return "";
  }
  PN*[] collect(str key)
  {
    PN*[] nl;
    PN* n=this.l;
    while (n)
    {
      if (n.r.rn==key)
        nl~=n;
      else
        if (n.r.rn=="")
          nl~=n.collect(key);
      n=n.n;
    }
    return nl;
  }
  T[] collectCall(str key)
  {
    T[] dl;
    PN* n=this.l;
    while (n)
    {
      if (n.r.rn==key)
        dl~=n.call();
      else
        if (n.r.rn=="")
          dl~=n.collectCall(key);
      n=n.n;
    }
    return dl;
  }
  str[] collectMatch(str key)
  {
    str[] sl;
    PN* n=this.l;
    while (n)
    {
      if (n.r.rn==key)
        sl~=n.match();
      else
        if (n.r.rn=="")
          sl~=n.collectMatch(key);
      n=n.n;
    }
    return sl;
  }
  void show(int lvl=0)
  {
    PN* n=this;
    str s;
    if (n.r)
      s=n.r.rn;
    else
      s="[?]";
    s~=":"~typeStr(n.r.rt);
    printf("%04i:%04i ",n.sb,n.se);
    printf("%s%s ",tsz(spaces(lvl)),tsz(s));
//    printf("p=%08p n=%08p l=%08p ",n,n.n,n.l);
    printf("n=%i l=%i ",n.n!=null,n.l!=null);
    printf("\"%s\" ",tsz(match()));
    printf("\n");
  }
  void showTree(int lvl=0)
  {
    this.show(lvl);
    PN* n=this.l;
    while (n)
    {
      if (n.l)
        n.l.showTree(lvl+1);
      n=n.n;
    }
  }
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- parse node allocator
//--------------------

public template PNAlloc(T)
{
  alias TParseNode!(T) PN;

  struct PNP
  {
    PN* ptr;
    int idx;
  }

  PNP 	freeNode; // free parse nodes

  // number of new nodes per allocation
  // values between 2^8 and 2^12 seem to work fine
  const int NNN=2048-4;

  PN* newPN()
  {
    // check if freelist needs refreshing
    if (!freeNode.idx)
    {
      freeNode.ptr=cast(PN*)(myalloc((PN).sizeof*NNN));
      freeNode.idx=NNN;
    }
    // get node from freelist
    return &(freeNode.ptr[--freeNode.idx]);
  }

  PN* dupPN(PN* o)
  {
    PN* n=newPN();
    *n=*o;
    return n;
  }
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- PseudoHash type
//--------------------

// pseudo hash type
//   what?
//     hash drop-in replacement
//   why?
//     fast iteration over elements
//   wtf?
//     while indexing by key is slow,
//     indexing by int is possible now

struct PseudoHash(T)
{
  T[] values;
  str[] keys;
	void clear()
	{
		keys.length=0;
		values.length=0;
	}
	bool exist(str key)
	{
		for (int k=0;k<keys.length;++k)
		{
			if (keys[k]==key)
				return true;
		}
		return false;
	}
	int idxByKey(str key)
	{
		for (int k=0;k<keys.length;++k)
		{
			if (keys[k]==key)
				return k;
		}
		throw new Exception(format("PseudoHash.idxByKey: cannot find key %s",key));
		return 0;
	}
	T* ptr(str key)
	{
		return &(values[this.idxByKey(key)]);
	}
	T opIndex(str key)
	{
		return values[this.idxByKey(key)];
	}
	void opIndexAssign(T val,str key)
	{
    try
    	values[this.idxByKey(key)]=val;
    catch (Exception e)
    {
      keys~=key;
      values~=val;
    }
	}
	T* ptr(size_t idx)
	{
		return &(values[idx]);
	}
	T opIndex(size_t idx)
	{
		return values[idx];
	}
	void opIndexAssign(T val,size_t idx)
	{
		values[idx]=val;
	}
	size_t length()
	{
		return keys.length;
	}
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- packrat parser class
//--------------------

/// Parser class;
class TDratParser(T)
{
  alias TRule!(T) Rule;
  alias TParseNode!(T) PN;
  alias T function(PN*) RuleFun;
  // parse node allocator
	mixin PNAlloc!(T);
	// uses void boxer
//	static mixin voidBoxer; // not any moer!
  // parsing rules
	PseudoHash!(Rule*) rule;
	// memo flag default
  bool mfDefault;
  // maximal source pointer during parse attempt
  int spmax;
  // magic node to signify fail in memo
  static PN* FAIL_NODE;
  // rule parser
  static TDratParser!(void*) ruleParser;
  // parse level
  const bool trackLevel=debugFlag;
  static if (trackLevel)
    int lvl=0;
  else
    const int lvl=0;
  // max parse level
  const bool saveMaxLevel=!true;
  static if (saveMaxLevel)
    int mxl=0;
  // default rule functions
  RuleFun[RuleType] defRuleFun;
  // the source to parse
  str src;      
  // source pointer
  int sp;       

  void useEmptyRuleFuns()
  {
    static T emptyRuleFun(PN* n)
    {
      throw new Exception("Uninitialised rule function called");
//      return T.init;
    }
    defRuleFun[RT.C]=&emptyRuleFun;
    defRuleFun[RT.S]=&emptyRuleFun;
    defRuleFun[RT.E]=&emptyRuleFun;
    defRuleFun[RT.R]=&emptyRuleFun;
    defRuleFun[RT.RN]=&emptyRuleFun;
    defRuleFun[RT.FST]=&emptyRuleFun;
    defRuleFun[RT.SEQ]=&emptyRuleFun;
    defRuleFun[RT.OOM]=&emptyRuleFun;
    defRuleFun[RT.ZOM]=&emptyRuleFun;
    defRuleFun[RT.OPT]=&emptyRuleFun;
    defRuleFun[RT.AND]=&emptyRuleFun;
    defRuleFun[RT.NOT]=&emptyRuleFun;
//    printf("emptyRuleFun=%p\n",&emptyRuleFun);
  }

  int min(int a,int b)
  {
    return (a<b)?a:b;
  }
  
  str deCtrl(str s)
  {
    str hex="0123456789abcdef";
    for (int k=0;k<s.length;)
    {
      if (s[k]<0x20)
      {
        str x;
        if (k>0)
          x~=s[0..k];
        x~="\\"~hex[s[k]>>4]~hex[s[k]&15];
        if (k<s.length)
          x~=s[k+1..$];
        s=x;
      }
      else
        ++k;
    }
    return s;
  }

  cptr sla(int sp)
  {
    int e=src.length;
    e=min(sp+20,e);
    return tsz(deCtrl("\""~src[sp..e]~"...\""));
  }

  int acceptChar(char c)
  {
    if (sp>=src.length)
      return 0;
    if (src[sp]!=c)
      return 0;
    ++sp;
    return 1;
  }

  int acceptStr(ccptr e)
  {
    if (sp>=src.length)
      return 0;
    ccptr s=cast(ccptr)(&(src[sp]));
    ccptr esave=e;
    while (*e)
      if (*(e++)!=*(s++))
        return 0;
    sp+=(e-esave);
    return 1;
  }

  int acceptRegexp(RegExp re)
  {
    foreach(m;re.search(src[sp..$]))
    {
      sp+=m.match(0).length;
      return 1;
    }
    return 0;
  }

  PN* accept(Rule* ru)
  {
    PN* res;
    if (ru.mf)
    {
      if (res=ru.me[sp],res)
      {
        if (res==FAIL_NODE)
        {
          return null;
        }
        else
        {
          sp=res.se;
          return dupPN(res);
        }
      }
    }
    if (ru.rn.length)
      dpf("%s%s:",tsz(spaces(lvl)),tsz(ru.rn));
    else
      dpf("%s:",tsz(spaces(lvl)));
    int sps=sp;
    if (sp>spmax) spmax=sp;
    static if (trackLevel)
    {
      ++lvl;
      static if (saveMaxLevel)
        if (lvl>mxl) mxl=lvl;
    }
  	switch (ru.rt)
  	{
			case RT.C   :
			{ // match character
				if (acceptChar(ru.c))
				{
				  res=newPN();
				  res.r=ru;
          res.n=null;
          res.l=null;
				  res.sb=sps;
				  res.se=sp;
				}
				dpf("C '%c' -> %p %s\n",ru.c,res,sla(sps));
				break;
			}
			case RT.S   :
			{ // match str
				if (acceptStr(ru.s.ptr))
				{
				  res=newPN();
				  res.r=ru;
          res.n=null;
          res.l=null;
				  res.sb=sps;
				  res.se=sp;
				}
				dpf("S '%s' -> %p %s\n",ru.rn.ptr,res,sla(sps));
				break;
			}
			case RT.E   :
			{ // match regexp
				if (acceptRegexp(ru.e))
				{
				  res=newPN();
				  res.r=ru;
          res.n=null;
          res.l=null;
				  res.sb=sps;
				  res.se=sp;
				}
				dpf("E '%s' -> %p %s\n",ru.rn.ptr,res,sla(sps));
				break;
			}
			case RT.FST :
			{ // match first
				dpf("FST\n");
				for (int k=0;k<ru.l.length;++k)
				{
				  sp=sps;
					res=accept(ru.l[k]);
          dpf("%sFST(%i) -> %i\n",tsz(spaces(lvl)),k,res);
					if (res)
					{
					  PN* r=newPN();
            r.r=ru;
            r.sb=sps;
            r.se=sp;
            r.n=null;
            r.l=res;
            res=r;
						break;
					}
				}
				break;
			}
			case RT.SEQ :
			{ // match all (serial)
				dpf("SEQ\n");
        PN* a,n;
        res=newPN();
        res.r=ru;
        res.sb=sps;
        res.n=null;
        res.l=null;
				for (int k=0;k<ru.l.length;++k)
				{
					n=accept(ru.l[k]);
          dpf("%sSEQ(%i) -> %i\n",tsz(spaces(lvl)),k,res);
					if (!n)
					{
					  res=null;
						break;
					}
					else
					{
					  if (res.l)
					    a.n=n;
					  else
					    res.l=n;
				    a=n;
					}
          res.se=sp;
				}
				dpf("%sSEQ -> %p\n",tsz(spaces(lvl)),res);
				break;
			}
			case RT.OPT :
			{ // option
				dpf("OPT\n");
			  res=newPN();
			  res.r=ru;
        res.n=null;
        res.l=accept(ru.r);
			  res.sb=sps;
			  res.se=sp;
				break;
			}
			case RT.ZOM :
			{ // zero-or-more
				dpf("ZOM\n");
			  res=newPN();
			  res.r=ru;
        res.n=null;
        res.l=accept(ru.r);
			  res.sb=sps;
        PN* a=res.l;
				while (sp>sps)
				{
				  sps=sp;
					a.n=accept(ru.r);
					a=a.n;
				}
				sps=res.sb;
			  res.se=sp;
				dpf("%sZOM -> %p\n",tsz(spaces(lvl)),res);
				break;
			}
			case RT.OOM :
			{ // one-or-more
				dpf("OOM\n");
			  res=newPN();
			  res.r=ru;
        res.n=null;
        res.l=accept(ru.r);
        if (!res.l)
        {
          res=null;
          break;
        }
			  res.sb=sps;
        PN* a=res.l;
				while (sp>sps)
				{
				  sps=sp;
					a.n=accept(ru.r);
					a=a.n;
				}
				sps=res.sb;
			  res.se=sp;
				break;
			}
			case RT.AND :
			{ // match this, don't consume input
				dpf("AND\n");
        res=accept(ru.r);
        if (res)
        {
				  res.r=ru;
          res.n=null;
          res.l=null;
				  res.sb=sps;
				  res.se=sps;
        }
        sp=sps;
				break;
			}
			case RT.NOT :
			{ // match all but this, don't consume input
				dpf("NOT\n");
        res=accept(ru.r);
        if (res)
        {
          res=null;
        }
        else
        {
				  res=newPN();
				  res.r=ru;
          res.n=null;
          res.l=null;
				  res.sb=sps;
				  res.se=sps;
        }
        sp=sps;
				break;
			}
			case RT.R :
			{ // match rule
        dpf("R\n");
        res=accept(ru.r);
        if (res)
        {
          if (ru.rn!="")
          {
//            printf("wrapping %s in %s\n",tsz(res.r.rn),tsz(ru.rn));
            PN* r=newPN();
            r.r=ru;
            r.n=null;
            r.l=res;
            r.sb=sps;
            r.se=sp;
            res=r;
          }
          else
          {
//            printf("unwrapping %s from %s\n",tsz(res.r.rn),tsz(ru.rn));
          }
        }
        break;
			}
			default      :
			{ // unknown
				throw new Exception(format("accept: unsupportet rule ",ru.rn," type ",ru.rt));
			}
    }
    if (ru.mf)
      if (res)
        ru.me[sps]=dupPN(res);
      else
        ru.me[sps]=FAIL_NODE;
    if (!res)
      sp=sps;
    static if (trackLevel)
      --lvl;
//    printf("%s%s@%i -> %p %s\n",tsz(spaces(lvl)),ru.rn.ptr,sps,res,las);
    return res;
  }

  void resolve(Rule** ru)
  { // resolve rules referenced by name
    (*ru).dp=this;
    (*ru).ps=&src;
  	switch ((*ru).rt)
  	{
			case RT.C   :
			case RT.S   :
			case RT.R   :
			case RT.E   :	break;
			case RT.RN  :
			{
        (*ru).rt=RT.R;
        (*ru).r=rule[*((*ru).n)];
				break;
			}
			case RT.OOM :
			case RT.ZOM :
			case RT.OPT :
			case RT.AND :
			case RT.NOT :
			{
				resolve(&((*ru).r));
				break;
			}
			case RT.FST :
			case RT.SEQ :
			{
				for (int k=0;k<(*ru).l.length;++k)
					resolve(&((*ru).l[k]));
				break;
			}
			default     :
			{
				throw new Exception("resolve: unsupportet rule type");
			}
  	}
  }
	
	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	//----------------------------------------------------------------------

  Rule* patchRules(Rule* ru)
  {
    ru.ps=&this.src;
    if (!ru.fu)
    {
      ru.fu=defRuleFun[ru.rt];
    }
  	switch (ru.rt)
  	{
			case RT.C   :
			case RT.S   :
			case RT.RN  :
			case RT.E   : break;
			case RT.R   :
			case RT.OOM :
			case RT.ZOM :
			case RT.OPT :
			case RT.AND :
			case RT.NOT :
			{
				patchRules(ru.r);
				break;
			}
			case RT.FST :
			case RT.SEQ :
			{
				for (int k=0;k<(*ru).l.length;++k)
					patchRules(ru.l[k]);
				break;
			}
			default     :
			{
				throw new Exception("patchRules: unsupportet rule type");
			}
  	}
  	return ru;
  }

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	//----------------------------------------------------------------------

  void patchRuleWS(Rule* ru)
  {
    // don't patch ws rule
    if (ru.rn=="ws")
      return ru;
    //
  	switch (ru.rt)
  	{
			case RT.C   :
			case RT.S   :
			case RT.R   :
			case RT.RN  :
			case RT.E   : break;
			case RT.OOM :
			case RT.ZOM :
			case RT.OPT :
			case RT.AND :
			case RT.NOT :
			{
				patchRuleWS(ru.r);
        switch (ru.r.rt)
        {
          case RT.C   :
          case RT.S   :
          case RT.E   : ru.r=Rule.newSEQ([ru.r,Rule.newR("ws")]);
          default     : {};
        }
				break;
			}
			case RT.SEQ :
			{
				for (int k=0;k<(*ru).l.length;++k)
					patchRuleWS(ru.l[k]);
        Rule*[] l=ru.l;
        for (int k=0;k<l.length;++k)
        {
          switch (l[k].rt)
          {
            case RT.C   :
            case RT.S   :
            case RT.E   :
            {
              Rule*[] newl;
              for (int kk=0;kk<l.length;++kk)
              {
                newl~=l[kk];
                if (kk==k)
                  newl~=Rule.newR("ws");
              }
              l=newl;
            }
            default     : {};
          }
        }
        ru.l=l;
				break;
			}
			case RT.FST :
			{
				for (int k=0;k<(*ru).l.length;++k)
					patchRuleWS(ru.l[k]);
        Rule*[] l=ru.l;
        for (int k=0;k<l.length;++k)
        {
          switch (l[k].rt)
          {
            case RT.C   :
            case RT.S   :
            case RT.E   : l[k]=Rule.newSEQ([l[k],Rule.newR("ws")]);
            default     : {};
          }
        }
        ru.l=l;
				break;
			}
			default     :
			{
				throw new Exception("patchRules: unsupportet rule type");
			}
  	}
  }

  void patchWS()
  {
		for (int k=0;k<rule.length;++k)
			patchRuleWS(rule[k]);
	}
	
	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	//----------------------------------------------------------------------

	void addRule(Rule* ru,RuleFun fu,bool mf)
	{
		ru.mf=mf;
		ru.fu=fu;
		if (rule.exist(ru.rn))
		  throw new Exception(format("rule '",ru.rn,"' does already exist"));
		rule[ru.rn]=patchRules(ru);
	}
	void addRule(Rule* ru,RuleFun fu)
	{
		addRule(ru,fu,mfDefault);
	}
	void addRule(Rule* ru,bool mf)
	{
		ru.mf=mf;
		rule[ru.rn]=patchRules(ru);
	}
	void addRule(Rule* ru)
	{
		addRule(ru,mfDefault);
	}
	void addRule(str rs,RuleFun fu,bool mf)
	{
		addRule(parseRule(rs),fu,mf);
	}
  void addRule(str rs,RuleFun fu)
  {
    addRule(parseRule(rs),fu);
  }
  void addRule(str rs,bool mf)
  {
    addRule(parseRule(rs),mf);
  }
  void addRule(str rs)
  {
    addRule(parseRule(rs));
  }

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	//----------------------------------------------------------------------

  static Rule* parseRule(str rs)
  {
    alias TParseNode!(void*) PN;
    alias TRule!(T) Rule;
    if (ruleParser is null)
      initRuleParser();
    // test
    PN* pt=ruleParser.parse("fst",rs);
    if (pt)
    {
      Rule* r=cast(Rule*)(pt.call());
      return r;
    }
    else
      throw new Exception("could not parse rule '"~rs~"'");
  }

  static void initRuleParser()
  {
    ruleParser=newRuleParser();
  }

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
  
  void showRules()
  {
		for (int k=0;k<rule.length;++k)
		{
		  Rule* r=rule[k];
		  r.show();
//		  writef("ptr=",r," rt=",r.rt," mf=",r.mf," me=",r.me.ptr," rn=",r.rn,"\n");
		}
  }
  
  int idx2lin(str s,int i)
  {
    int k,l;
    while (k<i)
    {
      if (s[k++]==0x0a)
        ++l;
    }
    return l;
  }
  
  void idx2lno(str s,int i,int *l,int *o)
  {
    int k;
    *l=1;
    *o=0;
    while (k<i)
    {
      if (s[k]==0x0a)
      {
        ++(*l);
        (*o)=i-k;
      }
      ++k;
    }
  }
  
  PN* parse(str r,str s,bool verbose=false)
  {
  	sp=0;
  	spmax=0;
    src=s~"\0";
    src.length=src.length-1;
//    printf("%s\n",tsz(src));
    //
		for (int k=0;k<rule.length;++k)
		{
		  // resolve named indirections
			resolve(rule.ptr(k));
      // allocate memos
		  if (rule[k].mf)
        rule[k].me=new PN*[src.length+1];
		}
    //
    if (verbose)
    {
      //showRules();
      printf("source length = %i\n",src.length);
      printf("start...\n");
    }
    d_time t=getUTCtime();
    //
  	PN* pt=accept(rule[r]);
  	int res=(pt!=null);
//    printf("%i/%i:%i\n",sp,src.length,res);
    //
    if (verbose)
      printf("res = %i\n",res);
    res=(sp>=(src.length));
    if (verbose)
    {
      printf("result = %i\n",res);
      printf("done after %f seconds.\n",(cast(double)(getUTCtime()-t))/TicksPerSecond);
    }
    if (!res)
    {
      int line=idx2lin(src,spmax)+1;
      printf("failed around line %i\n",line);
    }
    //
		for (int k=0;k<rule.length;++k)
		{
      // deallocate memos
      rule[k].me=null;
		}
    //
    if (res)
      return pt;
    else
      return null;
  }
  
  this(bool mf=true)
  {
    mfDefault=mf;
    if (!FAIL_NODE)
      FAIL_NODE=newPN();
    useEmptyRuleFuns();
  }
}

TDratParser!(void*) newRuleParser()
{
  alias void* T;
  alias TDratParser!(T) DratParser;
  alias TParseNode!(T) PN;
  alias TRule!(T) Rule;
  DratParser p=new DratParser();
  //
  useVptrRuleFuns(p);
  // fst
  p.addRule
  ( // ws seq ("/" ws seq)*
    Rule.newSEQ("fst",[
      Rule.newR("ws"),
      Rule.newR("seq"),
      Rule.newZOM(
        Rule.newSEQ([
          Rule.newE("[|/]"),
          Rule.newR("ws"),
          Rule.newR("seq")]))]),
    function T(PN* n)
    {
      dpf("[fst]\n");
      Rule*[] rl=cast(Rule*[])n.collectCall("seq");
      if (rl.length>1)
        return Rule.newFST(rl);
      else
        return rl[0];
    }
  );
  // seq
  p.addRule
  ( // seq:named+
    Rule.newOOM("seq",
      Rule.newR("named")),
    function T(PN* n)
    {
      dpf("[seq]\n");
      Rule*[] rl=cast(Rule*[])n.collectCall("named");
      if (rl.length>1)
        return Rule.newSEQ(rl);
      else
        return rl[0];
    }
  );
  // named
  p.addRule
  ( // name:$[_a-zA-Z][_a-zA-Z0-9]*:$? decor ws
    Rule.newSEQ("named",[
      Rule.newOPT(
        Rule.newE("id",r"[_a-zA-Z][-_a-zA-Z0-9]*:")),
      Rule.newR("ws"),
      Rule.newR("decor"),
      Rule.newR("ws")]),
    function T(PN* n)
    {
      dpf("[named]\n");
      void *d;
      // get data from decor
      d=n.findCall("decor");
      // name.id defined?
      if (PN* sn=n.find("id"))
      {
        str s=sn.match();
        // ###
        /*
        if ((cast(Rule*)d).rt==RT.RN)
        {
          (cast(Rule*)d).rt=RT.WN;
          (cast(Rule*)d).rn=s[0..$-1];
          //(cast(Rule*)d).fu=null;
          //d=cast(void*)Rule.newW(s[0..$-1],*((cast(Rule*)d).n));
        }
        else
        {
          (cast(Rule*)d).rn=s[0..$-1];
        }*/
        (cast(Rule*)d).rn=s[0..$-1];
      }
      return d;
    }
  );
  // decor
  p.addRule
  ( // prefix:$[&!]?$ undec postfix:$[?*+]?$ ws
    Rule.newSEQ("decor",[
      Rule.newE("prefix","[&!]?"),
      Rule.newR("undec"),
      Rule.newE("postfix","[?*+]?"),
      Rule.newR("ws")]),
    function T(PN* n)
    {
      dpf("[decor]\n");
      void *d;
      d=n.findCall("undec");
      str s;
      s=n.findMatch("postfix");
      if (s=="?")
        d=cast(T)Rule.newOPT(cast(Rule*)(d));
      else if (s=="*")
        d=cast(T)Rule.newZOM(cast(Rule*)(d));
      else if (s=="+")
        d=cast(T)Rule.newOOM(cast(Rule*)(d));
      s=n.findMatch("prefix");
      if (s=="&")
        d=cast(T)Rule.newAND(cast(Rule*)(d));
      else if (s=="!")
        d=cast(T)Rule.newNOT(cast(Rule*)(d));
      return d;
    }
  );
  // undec
  p.addRule
  ( // atom / ("(" ws opt ")" ws)
    Rule.newFST("undec",[
      Rule.newR("atom"),
      Rule.newSEQ([
        Rule.newC('('),
        Rule.newR("ws"),
        Rule.newR("fst"),
        Rule.newC(')'),
        Rule.newR("ws")])]),
    function T(PN* n)
    {
      dpf("[undec]\n");
      // subnode atom defined?
      if (PN* sn=n.find("atom"))
        return sn.call();
      // subnode brace defined?
      else
        return n.findCall("fst");
    }
  );
  // atom
  p.addRule
  ( // str / re / rule
    Rule.newFST("atom",[
      Rule.newR("str"),
      Rule.newR("re"),
      Rule.newR("rule")])
  );
  // rule
  p.addRule
  ( // expr:$[_a-zA-Z][_a-zA-Z0-9]*$ ws
    Rule.newSEQ("rule",[
      Rule.newE("expr",r"[_a-zA-Z][-_a-zA-Z0-9]*"),
      Rule.newR("ws")]),
    function T(PN* n)
    {
      str s=n.findMatch("expr");
      return cast(T)Rule.newR(s);
    }
  );
  // str
  p.addRule
  ( // expr:($'([^']|(\\'))*'$ / $"([^"]|(\\"))*"$) ws
    Rule.newSEQ("str",[
      Rule.newFST("expr",[
        Rule.newE(`'([^']|(\\'))*'`),
        Rule.newE(`"([^"]|(\\"))*"`)]),
      Rule.newR("ws")]),
    function T(PN* n)
    {
      str s=n.findMatch("expr");
      if (s.length==3)
        return cast(T)Rule.newC(s[1]);
      else
        return cast(T)Rule.newS(s[1..$-1]);
    }
  );
  // regexp
  p.addRule
  ( // "$" $([^$$]|($$$$))*$ "$" ws
    Rule.newFST("re",[
      Rule.newSEQ([
        Rule.newS("$"),
        Rule.newE("expr",`([^$]|(\$\$))*`),
        Rule.newS("$"),
        Rule.newR("ws")]),
      Rule.newSEQ([
        Rule.newS("%"),
        Rule.newE("expr",`([^%]|(\%\%))*`),
        Rule.newS("%"),
        Rule.newR("ws")])]),
    function T(PN* n)
    {
      return cast(T)Rule.newE(n.findMatch("expr"));
    }
  );
  // ws
  p.addRule
  ( // ($\s+$ / $#[^\r\n]*$)*
    Rule.newZOM("ws",
      Rule.newFST([
        Rule.newE(r"\s+"),
        Rule.newE(r"#[^\r\n]*")]))
  );
  //
  return p;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//--------------------
//-------------------- default rule function sets
//--------------------

void useVptrRuleFuns(TDratParser!(void*) p)
{
  p.defRuleFun[RT.C]=&vptrRuleFuns.funC;
  p.defRuleFun[RT.S]=&vptrRuleFuns.funS;
  p.defRuleFun[RT.E]=&vptrRuleFuns.funE;
  p.defRuleFun[RT.R]=&vptrRuleFuns.funR;
  p.defRuleFun[RT.FST]=&vptrRuleFuns.funFST;
  p.defRuleFun[RT.SEQ]=&vptrRuleFuns.funSEQ;
  p.defRuleFun[RT.OOM]=&vptrRuleFuns.funOOM;
  p.defRuleFun[RT.ZOM]=&vptrRuleFuns.funZOM;
  p.defRuleFun[RT.OPT]=&vptrRuleFuns.funOPT;
  p.defRuleFun[RT.AND]=&vptrRuleFuns.funAND;
  p.defRuleFun[RT.NOT]=&vptrRuleFuns.funNOT;
//  printf("vptrFST = %p\n",&vptrRuleFuns.funFST);
}

struct vptrRuleFuns
{
  alias void* T;
  alias TParseNode!(T) PN;
  
  static T funC(PN* n)
  {
    dpf2("C:%s\n",tsz(n.match()));
    str* s=(new str[1]).ptr;
    *s=n.match();
    return cast(T)(&(s[0]));
  };
  static T funS(PN* n)
  {
    dpf2("S:%s\n",tsz(n.match()));
    str* s=(new str[1]).ptr;
    *s=n.match();
    return cast(T)(s);
  };
  static T funE(PN* n)
  {
    dpf2("E:%s\n",tsz(n.match()));
    str* s=(new str[1]).ptr;
    *s=n.match();
    return cast(T)(s);
  };
  static T funR(PN* n)
  {
    dpf2("R:%s\n",tsz(n.match()));
    return n.l.call();
  };
  static T funFST(PN* n)
  {
    dpf2("FST:%s\n",tsz(n.match()));
    return n.l.call();
  };
  static T funSEQ(PN* n)
  {
    dpf2("SEQ:%s\n",tsz(n.match()));
    T[]* v=(new T[][1]).ptr;
    *v=n.subcalls();
    return cast(T)(v);
  };
  static T funOOM(PN* n)
  {
    dpf2("OOM:%s\n",tsz(n.match()));
    T[]* v=(new T[][1]).ptr;
    *v=n.subcalls();
    return cast(T)(v);
  };
  static T funZOM(PN* n)
  {
    dpf2("ZOM:%s\n",tsz(n.match()));
    T[]* v=(new T[][1]).ptr;
    *v=n.subcalls();
    return cast(T)(v);
  };
  static T funOPT(PN* n)
  {
    dpf2("OPT:%s\n",tsz(n.match()));
    if (n.l)
      return n.l.call();
    else
      return null;
  };
  static T funAND(PN* n)
  {
    dpf2("AND:%s\n",tsz(n.match()));
    return null;
  };
  static T funNOT(PN* n)
  {
    dpf2("NOT:%s\n",tsz(n.match()));
    return null;
  };
}

